/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifdef __GNUG__
#pragma implementation
#endif

#include "task.hxx"
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#if defined(__alpha) && !defined(__GNUG__)
#include "alphadecl.h"
#endif


TaskScheduler taskScheduler ;


void TaskScheduler::stackAlloc(int depth)
{

  char stack[TASK_STACK_SIZE] ;
  this->initStack(depth) ;
}


void TaskScheduler::initStack(int depth)
{
  if (depth < 0) {
    longjmp(this->env, 1) ;
  }
  
  if (setjmp(this->firstTask->env) != 0) {
    this->firstTask->start() ;
  }

  this->firstTask = this->firstTask->nextTask ;
  this->stackAlloc(depth-1) ; 
}


void TaskScheduler::start()
{
  if (setjmp(env) == 0) {      
    this->stackAlloc(this->runningTask-1) ;
  }

  this->resume(firstTask) ;
}


void TaskScheduler::schedule()
{
  while(1) {
    firstTask = firstTask->nextTask ;
    if (firstTask->status == TASK_RUNNING) {
      if (firstTask->delay == 0) {
	break ;
      } else {
	firstTask->delay-- ;
      }
    }
  }
}
 

void TaskScheduler::insert(Task* t) 
{
  if (this->firstTask != NULL) {
    t->nextTask = this->firstTask ;
    t->prevTask = this->firstTask->prevTask ;
    this->firstTask->prevTask->nextTask = t ;
    this->firstTask->prevTask = t ;
  } else {
    t->prevTask = t ;
    t->nextTask = t ;
  }
  this->firstTask = t ;
  this->setRunning(t) ;
}
 

void TaskScheduler::setIdle(Task* t) 
{
  t->status = TASK_IDLE ;
  this->runningTask-- ;
}


void TaskScheduler::setRunning(Task* t) 
{
  t->status = TASK_RUNNING ;
  this->runningTask++ ;
}


Task::Task() : delay(0)  
{
  taskScheduler.insert(this) ;
}


void Task::start()
{
}


void Task::yield(int cpt)
{
  this->delay = cpt ;
  if (this == taskScheduler.firstTask) {
    taskScheduler.schedule() ;
    if (this != taskScheduler.firstTask) {
      this->resume(taskScheduler.firstTask) ;
    }
  }
}


void Task::wait(PendingObject* obj)
{
  if (obj != NULL) {
    obj->waitingTask = this ;
  }
  taskScheduler.setIdle(this) ;
  this->yield() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Listener::Listener() : _numTask(0)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  for (int i = 0; i < TASK_LIST_MAX_FD; i++) {
    this->_waitingTask[i] = NULL ;
    this->_pset[i].fd = -1 ;
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
void Listener::start() 
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  while (1) {
    this->yield() ;     // Give up
    this->scanPoll() ;  // exec poll and activate transactions
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
void Listener::scanPoll()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (_numTask == 0) {
    return ;
  }
  
  // Look for an incoming event
  int res ;
  int tout = taskScheduler.active() > 1 ? 0 : -1 ;
  if ((res = poll(_pset, _numTask, tout)) < 0) {
    perror("==> Listener : unexpected error in poll") ;
    return ;
  } else if (res == 0) {
    return ;
  }

  // Notify activate transactions
  for (int i = 0; i < _numTask; i++) {
    if (_pset[i].revents != 0) {
      taskScheduler.setRunning(_waitingTask[i]) ;
      _pset[i].fd = -1 ;
    }
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
pollfd* Listener::taskRegister(ServerTask* task)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  for (int i = 0; _waitingTask[i] != NULL; i++) ;
  _waitingTask[i] = task ;
  _numTask = i+1 ;
  return &(_pset[i]) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
ServerTask::ServerTask(Listener* l, SvcServer* s) : _server(s)  
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->_afd = l->taskRegister(this) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void ServerTask::start()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  while (1) {
    if (_server->fd() != -1) {
      this->_afd->fd = _server->fd() ;
    } else {
      this->_afd->fd = _server->listenFd() ;
    }
    this->_afd->events = POLLIN ;
    this->wait() ;
    // Respond to the event
    this->replyEvent() ;
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
SvcEvent ServerTask::findEvent()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (_server->isConnected() == 0) {
    return SVC_CONNECT ;
  }

  int dummy ;
  int size = recv(_server->fd(), (char*)&dummy, 4, MSG_PEEK) ;

  if (size == 0) {
    return SVC_DISCONNECT ;
  } else {
    return SVC_DATA ;
  } 
}


/*= EXTERNAL FUNCTION =======================================================*/
void ServerTask::replyEvent()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  SvcEvent ev = findEvent() ;

  switch (ev) {
  case SVC_CONNECT :
    this->_server->accept() ;
    break ;
  case SVC_DATA :
    this->_server->execRpc() ;
    while (this->_server->inEmpty() == FALSE) {
      this->yield() ;
      this->_server->execRpc() ;
    } 
    break ;
  case SVC_DISCONNECT :
    this->lostConnection() ;    
    break ;
  }  
}


/*= EXTERNAL FUNCTION =======================================================*/
int ServerTask::open(char* name, int flag, int mode)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return ::open(name, flag | O_NONBLOCK, mode) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int ServerTask::lseek(int fd, off_t offset, int whence)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return ::lseek(fd, offset, whence) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int ServerTask::read(int fd, char* buffer, int nbytes)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return ::read(fd, buffer, nbytes) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int ServerTask::write(int fd, char* buffer, int nbytes)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return ::write(fd, buffer, nbytes) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void ServerTask::lostConnection()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->_server->lostConnection() ;
  this->_afd->fd = _server->listenFd() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void taskWaitIO(int fd, int events)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (taskScheduler.firstTask == NULL) return ;
  ((ServerTask *)taskScheduler.firstTask)->_afd->fd = fd ;
  ((ServerTask *)taskScheduler.firstTask)->_afd->events = events ;
  taskScheduler.firstTask->wait() ;
}

